#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-12-23
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import pathlib

from sdsstools import get_config, get_package_version


# pip package name
NAME = "sdss-lvmecp"

# Loads config. config name is the package name.
config = get_config(
    "lvmbeat",
    config_file=pathlib.Path(__file__).parent / "config.yaml",
)


# package name should be pip package name
__version__ = get_package_version(path=__file__, package_name=NAME)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-12-20
# @Filename: heartbeats.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import time

from typing import NotRequired, TypedDict


class HeartbeatData(TypedDict):
    """Data to create a heartbeat."""

    name: str
    critical: NotRequired[bool]


class Heartbeats(dict[str, "Heartbeat"]):
    """Collection of heartbeats."""

    def __init__(self, data: list[HeartbeatData] = []):
        dict.__init__({})
        for heartbeat_data in data:
            self[heartbeat_data["name"]] = Heartbeat(
                heartbeat_data["name"],
                critical=heartbeat_data.get("critical", True),
            )

    def seen_since(self, delta: float = 30):
        """Returns which components have been seen in the last ``delta`` seconds."""

        response: dict[str, bool] = {}

        for key, heartbeat in self.items():
            heartbeat_delta = heartbeat.time_delta()
            response[key] = heartbeat_delta is not None and heartbeat_delta < delta

        return response


class Heartbeat:
    """Tracks and updates the heartbeat of a component."""

    def __init__(self, name: str, critical: bool = True):
        self.name = name
        self.critical = critical
        self.last_set: float | None = None

    def set(self):
        """Sets the heartbeat to the current time."""

        self.last_set = time.time()

    def time_delta(self):
        """Returns the number of seconds since the last heartbeat."""

        if self.last_set is None:
            return None

        return time.time() - self.last_set



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-12-23
# @Filename: __main__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import os
from copy import deepcopy

import click
from click_default_group import DefaultGroup

from sdsstools import Configuration
from sdsstools.daemonizer import DaemonGroup, cli_coro

from lvmbeat import config
from lvmbeat.actor import BeatActor


@click.group(cls=DefaultGroup, default="actor", default_if_no_args=True)
@click.option(
    "-c",
    "--config",
    "config_file",
    type=click.Path(exists=True, dir_okay=False),
    help="Path to the user configuration file.",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    help="Debug mode. Use additional v for more details.",
)
@click.pass_context
def lvmbeat(ctx, verbose, config_file: str | None):
    """Heartbeat middleware and internet connection status."""

    ctx.obj = {
        "verbose": verbose,
        "config_file": config_file,
    }


@lvmbeat.group(cls=DaemonGroup, prog="beat-actor", workdir=os.getcwd())
@click.option(
    "--with-simulator",
    is_flag=True,
    help="Runs the actor aginst the simulator.",
)
@click.pass_context
@cli_coro()
async def actor(ctx, with_simulator: bool = False):
    """Runs the actor."""

    cli_config_file = ctx.obj["config_file"]
    if cli_config_file:
        beat_config = Configuration(cli_config_file, base_config=config)
        print(f"Using config file {cli_config_file}")
    else:
        beat_config = deepcopy(config)
        print("Using internal configuration.")

    actor_obj = BeatActor.from_config(beat_config)
    config.update(beat_config)

    await actor_obj.start()
    await actor_obj.run_forever()


def main():
    lvmbeat(auto_envvar_prefix="LVMBEAT")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-12-22
# @Filename: monitor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import logging
import os
import time
from contextlib import asynccontextmanager
from dataclasses import dataclass, field

import fastapi_utils
import fastapi_utils.tasks
from fastapi import FastAPI, HTTPException
from fastapi.datastructures import State
from lvmopstools.notifications import send_critical_error_email

from lvmbeat import __version__, config


logger = logging.getLogger("uvicorn.error")


@dataclass
class EmailSettings:
    """A class to store the state of the monitor webapp."""

    recipients: list[str] = field(default_factory=list)
    from_address: str | None = None
    email_reply_to: str | None = None
    host: str | None = None
    port: int | None = None
    tls: bool = False
    username: str | None = None
    password: str | None = None


@asynccontextmanager
async def lifespan(app: FastAPI):
    """Handles the application life-cycle."""

    await check_heartbeat()

    yield


app = FastAPI(swagger_ui_parameters={"tagsSorter": "alpha"}, lifespan=lifespan)
app.state = State(
    {
        "active": False,
        "last_seen": time.time(),
        "enabled": True,
    }
)


def get_email_settings():
    """Returns the email settings reading from environment variables."""

    recipients_envvar = os.getenv("LVMBEAT_EMAIL_RECIPIENTS", [])
    if isinstance(recipients_envvar, str):
        recipients = recipients_envvar.split(",")
    else:
        recipients = []

    from_address = os.getenv("LVMBEAT_EMAIL_FROM_ADDRESS")
    email_reply_to = os.getenv("LVMBEAT_EMAIL_REPLY_TO")
    host = os.getenv("LVMBEAT_EMAIL_HOST")

    port_envvar = os.getenv("LVMBEAT_EMAIL_PORT")
    port = int(port_envvar) if port_envvar else None

    tls = os.getenv("LVMBEAT_EMAIL_TLS")
    if tls and (tls is True or tls.lower() in ["true", "yes", "1"]):
        tls = True
    else:
        tls = False

    username = os.getenv("LVMBEAT_EMAIL_USERNAME")
    password = os.getenv("LVMBEAT_EMAIL_PASSWORD")

    return EmailSettings(
        recipients=recipients,
        from_address=from_address,
        email_reply_to=email_reply_to or from_address,
        host=host,
        port=port,
        tls=tls,
        username=username,
        password=password,
    )


def send_email(message: str, subject: str):
    """Sends a critical alert email notification."""

    email_settings = get_email_settings()

    if len(email_settings.recipients) == 0:
        raise HTTPException(status_code=400, detail="No recipients defined.")
    if email_settings.from_address is None:
        raise HTTPException(status_code=400, detail="No from_address defined.")
    if email_settings.host is None:
        raise HTTPException(status_code=400, detail="No host defined.")
    if email_settings.port is None:
        raise HTTPException(status_code=400, detail="No port defined.")
    if email_settings.tls:
        if email_settings.username is None:
            raise HTTPException(status_code=400, detail="No username defined.")
        if email_settings.password is None:
            raise HTTPException(status_code=400, detail="No password defined.")

    send_critical_error_email(
        message=message,
        subject=subject,
        recipients=email_settings.recipients,
        from_address=email_settings.from_address,
        email_reply_to=email_settings.email_reply_to,
        host=email_settings.host,
        port=email_settings.port,
        tls=email_settings.tls,
        username=email_settings.username,
        password=email_settings.password,
    )


@fastapi_utils.tasks.repeat_every(seconds=20, logger=logger, raise_exceptions=False)
def check_heartbeat():
    """Checks if we have received a heartbeat from LCO or sends an alert."""

    logger.debug("Checking heartbeat.")

    if not app.state.enabled:
        logger.debug("Heartbeat monitor is disabled.")
        return

    max_time_to_alert: float = float(
        os.getenv(
            "LVMBEAT_SEND_EMAIL_AFTER",
            config["outside_monitor.send_email_after"],
        )
    )

    now = time.time()
    if not app.state.active and now - app.state.last_seen > max_time_to_alert:
        logger.warning(
            f"No heartbeat received in the last {max_time_to_alert} seconds. "
            "Sending critical alert email."
        )
        send_email(
            message="The LCO internet connection is down.",
            subject="LCO internet is down",
        )
        app.state.active = True

    elif app.state.active and now - app.state.last_seen < max_time_to_alert:
        logger.info("Heartbeat received. Resetting alert and sending all-clear email.")
        send_email(
            message="The LCO internet connection appears to be up.",
            subject="RESOLVED: LCO internet is up",
        )
        app.state.active = False


@app.get("/heartbeat", description="Sets the heartbeat.")
def route_get_heartbeat():
    """Sets the heartbeat."""

    app.state.last_seen = time.time()

    return {"message": "Heartbeat received."}


@app.get("/heartbeat/status", description="Status of the heartbeat monitor.")
def route_get_heartbeat_status():
    """Status of the heartbeat monitor."""

    return {
        "enabled": app.state.enabled,
        "active": app.state.active,
        "last_seen": app.state.last_seen,
    }


@app.get("/heartbeat/enable", description="Enables the heartbeat monitor.")
def route_get_heartbeat_enable():
    """Enables the heartbeat monitor."""

    app.state.enabled = True

    return {"message": "Heartbeat monitor enabled."}


@app.get("/heartbeat/disable", description="Disables the heartbeat monitor.")
def route_get_heartbeat_disable():
    """Disables the heartbeat monitor."""

    app.state.enabled = False

    return {"message": "Heartbeat monitor disabled."}


@app.get("/email/test", description="Sends a test email.")
def route_get_email_test():
    """Sends a test email."""

    send_email("This is a test message. Please ignore.", "TEST: LCO internet is down")

    return {"message": "Email sent."}


@app.get("/version", description="Returns the version of the monitor.")
def route_get_version():
    """Returns the version of the monitor."""

    return {"version": __version__}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-12-21
# @Filename: tools.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from datetime import datetime, timezone

from nmap3 import NmapHostDiscovery

from sdsstools.utils import run_in_executor


__all__ = ["timestamp_to_iso", "is_host_up"]


def timestamp_to_iso(ts: float | None) -> str | None:
    """Converts a timestamp to an ISO string."""

    if ts is None:
        return None

    return datetime.fromtimestamp(ts, timezone.utc).isoformat().replace("+00:00", "Z")


async def is_host_up(host: str) -> bool:
    """Returns whether a host is up.

    Parameters
    ----------
    host
        The host to check.

    Returns
    -------
    is_up
        ``True`` if the host is up, ``False`` otherwise.

    """

    nmap = NmapHostDiscovery()
    result = await run_in_executor(
        nmap.nmap_no_portscan,
        host,
        args="--host-timeout=1 --max-retries=2",
    )

    if (
        host not in result
        or "state" not in result[host]
        or "state" not in result[host]["state"]
    ):
        return False

    return result[host]["state"]["state"] == "up"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-12-20
# @Filename: actor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
from time import time

from typing import Annotated, TypedDict

import click
import httpx
from pydantic import BaseModel, Field

from clu.actor import AMQPActor
from clu.command import Command
from clu.parsers.click import command_parser
from sdsstools import cancel_task

from lvmbeat import config
from lvmbeat.heartbeats import HeartbeatData, Heartbeats
from lvmbeat.tools import is_host_up, timestamp_to_iso


class HeartbeatSchema(BaseModel):
    """Schema for a heartbeat."""

    current: Annotated[
        bool,
        Field(description="Whether the heartbeat has been seen and is active."),
    ]
    last_seen: Annotated[
        str | None,
        Field(description="ISO date of the last time the heartbeat was seen."),
    ]


class BeatKeywordSchema(BaseModel):
    """Reply schema for :obj:`.BeatActor`."""

    heartbeats: Annotated[
        dict[str, HeartbeatSchema],
        Field(description="Status of the component heartbeats."),
    ]
    last_emitted_ecp: Annotated[
        str | None,
        Field(description="ISO date of the last hearbeat sent to the ECP."),
    ]
    last_emitted_outside: Annotated[
        str | None,
        Field(
            description="ISO date of the last successful "
            "heartbeat sent to the outside world monitoring tool."
        ),
    ]
    network: Annotated[
        NetworkStatus,
        Field(description="Network status."),
    ]


class NetworkStatus(TypedDict):
    """Network status."""

    lco: bool
    outside: bool


class BeatActor(AMQPActor):
    """Heartbeat actor."""

    parser = command_parser

    def __init__(self, *args, heartbeats: list[HeartbeatData] = [], **kwargs):
        if "schema" not in kwargs:
            kwargs["schema"] = BeatKeywordSchema

        super().__init__(*args, **kwargs)

        # Number of seconds after which a heartbeat is considered to be down.
        self.timeout: float = self.config.get("timeout", 30)

        # Create the heartbeats instance. The information is usually passed
        # in the configuration as the heartbeats key.
        heartbeat_data = heartbeats or self.config.get("heartbeats", [])
        self.heartbeats = Heartbeats(heartbeat_data)

        # Last heartbeat emitted to the ECP
        self._last_emitted_ecp: float | None = None

        # Last time we emitted the outside world heartbeat.
        self._last_emitted_outside: float | None = None

        # Track network access to other LCO services and the outside world.
        self.network_status: NetworkStatus = {
            "lco": True,
            "outside": True,
        }

        self._emit_outside_task = asyncio.create_task(self.emit_outside())
        self._network_status_task = asyncio.create_task(self.update_network_status())

    async def stop(self):
        """Stops the actor."""

        self._emit_outside_task = await cancel_task(self._emit_outside_task)

        return await super().stop()

    async def update(self):
        """Callback called when a heartbeat gets updated.

        If all the critical heartbeats are present, emits the dome heartbeat.

        """

        seen_hbs = self.heartbeats.seen_since(self.timeout)
        for hb_name, seen in seen_hbs.items():
            hb = self.heartbeats[hb_name]
            if seen is False:
                if hb.critical is True:
                    self.log.warning(
                        f"Heartbeat for {hb_name!r} has not been seen "
                        f"in the last {self.timeout} seconds."
                    )
                    return
                else:
                    self.log.warning(
                        f"Heartbeat for {hb_name!r} has not been seen "
                        f"in the last {self.timeout} seconds. "
                        "Skipping since it is not critical."
                    )

        if self._last_emitted_ecp and time() - self._last_emitted_ecp < 10:
            # Prevent emitting the heartbeat too often.
            return

        self.log.info("Emitting dome heartbeat.")
        cmd = await self.send_command("lvmecp", "heartbeat")
        if not cmd.status.did_succeed:
            self.log.error(
                "Possible error emitting dome heartbeat. "
                "The lcmecp heartbeat command failed."
            )

    async def emit_outside(self):
        """Emits a heartbeat to the outside world."""

        outside_url = config["outside_monitor"]["url"]
        if not outside_url:
            self.log.warning(
                "No outside monitor URL defined. Will not emit "
                "heartbeats to the outside world."
            )
            return

        while True:
            async with httpx.AsyncClient(timeout=5, follow_redirects=True) as client:
                try:
                    response = await client.get(outside_url)
                    response.raise_for_status()
                except Exception as ee:
                    self.log.error(f"Error emitting heartbeat to {outside_url!r}: {ee}")
                else:
                    self.log.debug(f"Emitted heartbeat to {outside_url!r}.")
                    self._last_emitted_outside = time()

            await asyncio.sleep(15)

    async def update_network_status(self):
        """Updates the network status."""

        while True:
            self.log.debug("Updating network status.")

            internet = await is_host_up("8.8.8.8")  # Google DNS
            lco = await is_host_up("10.8.8.46")  # clima.lco.cl

            self.network_status["outside"] = internet
            self.network_status["lco"] = lco

            await asyncio.sleep(15)


BeatCommand = Command[BeatActor]


@command_parser.command()
async def status(command: BeatCommand):
    """Outputs the status of the heartbeats."""

    seen = command.actor.heartbeats.seen_since(command.actor.timeout)
    last_emitted_ecp = timestamp_to_iso(command.actor._last_emitted_ecp)
    last_emitted_outside = timestamp_to_iso(command.actor._last_emitted_outside)

    heartbeats = {
        hb_name: {
            "current": seen[hb_name],
            "last_seen": timestamp_to_iso(hb.last_set),
        }
        for hb_name, hb in command.actor.heartbeats.items()
    }

    return command.finish(
        heartbeats=heartbeats,
        last_emitted_ecp=last_emitted_ecp,
        last_emitted_outside=last_emitted_outside,
        network=command.actor.network_status,
    )


@command_parser.command()
@click.argument("heartbeat", type=str)
async def set(command: BeatCommand, heartbeat: str):
    """Sets the heartbeat for a component."""

    if heartbeat not in command.actor.heartbeats:
        return command.fail(f"Heartbeat {heartbeat!r} not found.")

    command.actor.heartbeats[heartbeat].set()
    await command.actor.update()

    return command.finish(text="Heartbeat set.")
